@@ -11,6 +11,7 @@
 #include <sbkpep.h>
 
 #include <QtQuick/QQuickPaintedItem>
+#include <QtQuick/QQuickTextDocument>
 
 #if QT_CONFIG(opengl) || QT_CONFIG(opengles2) || QT_CONFIG(opengles3)
 #  include <QtQuick/QQuickFramebufferObject>
@@ -73,6 +74,7 @@ void PySide::initQuickSupport(PyObject *module)
     qRegisterMetaType<QQuickFramebufferObject*>("QQuickFramebufferObject*");
 #endif
     qRegisterMetaType<QQuickItem*>("QQuickItem*");
+    qRegisterMetaType<QQuickTextDocument*>("QQuickTextDocument*");
 
     Qml::setQuickRegisterItemFunction(quickRegisterType);
 }
