@@ -28,21 +28,24 @@
 #include <basewrapper.h>
 #include <bindingmanager.h>
 #include <gilstate.h>
+#include <helper.h>
 #include <sbkconverter.h>
+#include <sbkerrors.h>
+#include <sbkpep.h>
 #include <sbkstring.h>
 #include <sbkstaticstrings.h>
 #include <sbkfeature_base.h>
 #include <sbkmodule.h>
 
-#include <QtCore/QByteArray>
-#include <QtCore/QCoreApplication>
-#include <QtCore/QDebug>
-#include <QtCore/QDir>
-#include <QtCore/QFileInfo>
-#include <QtCore/QMetaMethod>
-#include <QtCore/QMutex>
-#include <QtCore/QStack>
-#include <QtCore/QThread>
+#include <QtCore/qbytearray.h>
+#include <QtCore/qcoreapplication.h>
+#include <QtCore/qdebug.h>
+#include <QtCore/qdir.h>
+#include <QtCore/qfileinfo.h>
+#include <QtCore/qmetaobject.h>
+#include <QtCore/qmutex.h>
+#include <QtCore/qstack.h>
+#include <QtCore/qthread.h>
 #include <QtCore/private/qobject_p.h>
 
 #include <algorithm>
@@ -55,14 +58,17 @@
 #ifdef Q_OS_WIN
 #  include <conio.h>
 #else
-#  include <QtCore/QDeadlineTimer>
+#  include <QtCore/qdeadlinetimer.h>
 #  include <QtCore/private/qcore_unix_p.h>
 #endif
 
 using namespace Qt::StringLiterals;
 
 static QStack<PySide::CleanupFunction> cleanupFunctionList;
-static void *qobjectNextAddr;
+
+// Used by QML (main thread), but needs to be protected against other
+// threads constructing QObject's.
+static void thread_local *qobjectNextAddr;
 
 QT_BEGIN_NAMESPACE
 extern bool qRegisterResourceData(int, const unsigned char *, const unsigned char *,
@@ -409,7 +415,8 @@ static void destructionVisitor(SbkObject *pyObj, void *data)
     auto *pyQApp = reinterpret_cast<SbkObject *>(realData[0]);
     auto *pyQObjectType = reinterpret_cast<PyTypeObject *>(realData[1]);
 
-    if (pyObj != pyQApp && PyObject_TypeCheck(pyObj, pyQObjectType)) {
+    auto *ob = reinterpret_cast<PyObject *>(pyObj);
+    if (pyObj != pyQApp && PyObject_TypeCheck(ob, pyQObjectType)) {
         if (Shiboken::Object::hasOwnership(pyObj) && Shiboken::Object::isValid(pyObj, false)) {
             Shiboken::Object::setValidCpp(pyObj, false);
 
@@ -451,15 +458,13 @@ std::size_t getSizeOfQObject(PyTypeObject *type)
     return retrieveTypeUserData(type)->cppObjSize;
 }
 
-void initDynamicMetaObject(PyTypeObject *type, const QMetaObject *base, std::size_t cppObjSize)
+static void initDynamicMetaObjectHelper(PyTypeObject *type,
+                                        TypeUserData *userData)
 {
-    //create DynamicMetaObject based on python type
-    auto *userData = new TypeUserData(reinterpret_cast<PyTypeObject *>(type), base, cppObjSize);
-    userData->mo.update();
     Shiboken::ObjectType::setTypeUserData(type, userData, Shiboken::callCppDestructor<TypeUserData>);
 
-    //initialize staticQMetaObject property
-    void *metaObjectPtr = const_cast<QMetaObject *>(userData->mo.update());
+    // initialize staticQMetaObject property
+    const void *metaObjectPtr = userData->mo.update();
     static SbkConverter *converter = Shiboken::Conversions::getConverter("QMetaObject");
     if (!converter)
         return;
@@ -468,6 +473,11 @@ void initDynamicMetaObject(PyTypeObject *type, const QMetaObject *base, std::siz
                      Shiboken::PyName::qtStaticMetaObject(), pyMetaObject);
 }
 
+void initDynamicMetaObject(PyTypeObject *type, const QMetaObject *base, std::size_t cppObjSize)
+{
+    initDynamicMetaObjectHelper(type, new TypeUserData(base, cppObjSize));
+}
+
 TypeUserData *retrieveTypeUserData(PyTypeObject *pyTypeObj)
 {
     if (!SbkObjectType_Check(pyTypeObj))
@@ -519,7 +529,9 @@ void initQObjectSubType(PyTypeObject *type, PyObject *args, PyObject * /* kwds *
     // PYSIDE-1463: Don't change feature selection durin subtype initialization.
     // This behavior is observed with PySide 6.
     PySide::Feature::Enable(false);
-    initDynamicMetaObject(type, userData->mo.update(), userData->cppObjSize);
+    // create DynamicMetaObject based on python type
+    auto *subTypeData = new TypeUserData(type, userData->mo.update(), userData->cppObjSize);
+    initDynamicMetaObjectHelper(type, subTypeData);
     PySide::Feature::Enable(true);
 }
 
@@ -594,10 +606,7 @@ PyObject *getHiddenDataFromQObject(QObject *cppSelf, PyObject *self, PyObject *n
 
     // Search on metaobject (avoid internal attributes started with '__')
     if (!attr) {
-        PyObject *type{};
-        PyObject *value{};
-        PyObject *traceback{};
-        PyErr_Fetch(&type, &value, &traceback);     // This was omitted for a loong time.
+        Shiboken::Errors::Stash errorStash;
 
         int flags = currentSelectId(Py_TYPE(self));
         int snake_flag = flags & 0x01;
@@ -622,8 +631,10 @@ PyObject *getHiddenDataFromQObject(QObject *cppSelf, PyObject *self, PyObject *n
                     if (res) {
                         AutoDecRef elemName(PyObject_GetAttr(res, PySideMagicName::name()));
                         // Note: This comparison works because of interned strings.
-                        if (elemName == name)
+                        if (elemName == name) {
+                            errorStash.release();
                             return res;
+                        }
                         Py_DECREF(res);
                     }
                     PyErr_Clear();
@@ -654,25 +665,26 @@ PyObject *getHiddenDataFromQObject(QObject *cppSelf, PyObject *self, PyObject *n
                     } else if (auto *func = MetaFunction::newObject(cppSelf, i)) {
                         auto *result = reinterpret_cast<PyObject *>(func);
                         PyObject_SetAttr(self, name, result);
+                        errorStash.release();
                         return result;
                     }
                 }
             }
             if (!signalList.isEmpty()) {
                 auto *pySignal = reinterpret_cast<PyObject *>(
-                    Signal::newObjectFromMethod(self, signalList));
+                    Signal::newObjectFromMethod(cppSelf, self, signalList));
                 PyObject_SetAttr(self, name, pySignal);
+                errorStash.release();
                 return pySignal;
             }
         }
-        PyErr_Restore(type, value, traceback);
     }
     return attr;
 }
 
 bool inherits(PyTypeObject *objType, const char *class_name)
 {
-    if (strcmp(objType->tp_name, class_name) == 0)
+    if (std::strcmp(PepType_GetFullyQualifiedNameStr(objType), class_name) == 0)
         return true;
 
     PyTypeObject *base = objType->tp_base;
@@ -717,11 +729,14 @@ static void invalidatePtr(any_t *object)
     if (Py_IsInitialized() == 0)
         return;
 
-    Shiboken::GilState state;
-
-    SbkObject *wrapper = Shiboken::BindingManager::instance().retrieveWrapper(object);
-    if (wrapper != nullptr)
-        Shiboken::BindingManager::instance().releaseWrapper(wrapper);
+    // Check for existence before locking (fix hang when QObjects
+    // are moved to different threads).
+    auto &bindingManager = Shiboken::BindingManager::instance();
+    if (bindingManager.hasWrapper(object)) {
+        Shiboken::GilState state;
+        if (SbkObject *wrapper = bindingManager.retrieveWrapper(object))
+            bindingManager.releaseWrapper(wrapper);
+    }
 }
 
 static const char invalidatePropertyName[] = "_PySideInvalidatePtr";
@@ -779,10 +794,7 @@ PyTypeObject *getTypeForQObject(const QObject *cppSelf)
     if (existing != nullptr)
         return reinterpret_cast<PyObject *>(existing)->ob_type;
     // Find the best match (will return a PySide type)
-    auto *sbkObjectType = Shiboken::ObjectType::typeForTypeName(typeName(cppSelf));
-    if (sbkObjectType != nullptr)
-        return reinterpret_cast<PyTypeObject *>(sbkObjectType);
-    return nullptr;
+    return Shiboken::ObjectType::typeForTypeName(typeName(cppSelf));
 }
 
 PyObject *getWrapperForQObject(QObject *cppSelf, PyTypeObject *sbk_type)
@@ -875,9 +887,7 @@ QString pyPathToQString(PyObject *path)
 
 bool isCompiledMethod(PyObject *callback)
 {
-    return PyObject_HasAttr(callback, PySide::PySideName::im_func())
-           && PyObject_HasAttr(callback, PySide::PySideName::im_self())
-           && PyObject_HasAttr(callback, PySide::PySideMagicName::code());
+    return Shiboken::isCompiledMethod(callback);
 }
 
 static const unsigned char qt_resource_name[] = {
@@ -927,7 +937,7 @@ bool registerInternalQtConf()
     // This will disable the internal qt.conf which points to the PySide6 subdirectory (due to the
     // subdirectory not existing anymore).
 #ifndef PYPY_VERSION
-    QString executablePath = QString::fromWCharArray(Py_GetProgramFullPath());
+    QString executablePath = QString::fromWCharArray(PyUnicode_AsWideCharString(PySys_GetObject("executable"), nullptr));
 #else
     // PYSIDE-535: FIXME: Add this function when available.
     QString executablePath = QLatin1StringView("missing Py_GetProgramFullPath");
@@ -1059,10 +1069,10 @@ QMetaType qMetaTypeFromPyType(PyTypeObject *pyType)
         return QMetaType(QMetaType::Int);
     if (Shiboken::ObjectType::checkType(pyType))
         return QMetaType::fromName(Shiboken::ObjectType::getOriginalName(pyType));
-    return QMetaType::fromName(pyType->tp_name);
+    return QMetaType::fromName(PepType_GetFullyQualifiedNameStr(pyType));
 }
 
-debugPyTypeObject::debugPyTypeObject(const PyTypeObject *o) noexcept
+debugPyTypeObject::debugPyTypeObject(PyTypeObject *o) noexcept
     : m_object(o)
 {
 }
@@ -1074,7 +1084,7 @@ QDebug operator<<(QDebug debug, const debugPyTypeObject &o)
     debug.nospace();
     debug << "PyTypeObject(";
     if (o.m_object)
-        debug << '"' << o.m_object->tp_name << '"';
+        debug << '"' << PepType_GetFullyQualifiedNameStr(o.m_object) << '"';
     else
         debug << '0';
     debug << ')';
@@ -1211,10 +1221,6 @@ QDebug operator<<(QDebug debug, const debugPyObject &o)
     return debug;
 }
 
-debugPyBuffer::debugPyBuffer(Py_buffer *b) noexcept : m_buffer(b)
-{
-}
-
 static void formatPy_ssizeArray(QDebug &debug, const char *name, const Py_ssize_t *array, int len)
 {
     debug << ", " << name << '=';
@@ -1228,6 +1234,11 @@ static void formatPy_ssizeArray(QDebug &debug, const char *name, const Py_ssize_
     }
 }
 
+#if !defined(Py_LIMITED_API) || Py_LIMITED_API >= 0x030B0000
+debugPyBuffer::debugPyBuffer(Py_buffer *b) noexcept : m_buffer(b)
+{
+}
+
 PYSIDE_API QDebug operator<<(QDebug debug, const debugPyBuffer &b)
 {
     QDebugStateSaver saver(debug);
@@ -1253,5 +1264,6 @@ PYSIDE_API QDebug operator<<(QDebug debug, const debugPyBuffer &b)
     debug << ')';
     return debug;
 }
+#endif // !Py_LIMITED_API || >= 3.11
 
 } // namespace PySide
