@@ -11,7 +11,8 @@
 #include "typesystem_typedefs.h"
 #include "typesystem_enums.h"
 
-#include <QtCore/QRegularExpression>
+#include <QtCore/qhash.h>
+#include <QtCore/qregularexpression.h>
 
 #include <array>
 #include <optional>
@@ -82,6 +83,7 @@ public:
     };
     Q_DECLARE_FLAGS(AttroCheck, AttroCheckFlag);
 
+    using FunctionMapping = QHash<AbstractMetaFunctionCPtr, AbstractMetaFunctionCPtr>;
     using FunctionGroups = QMap<QString, AbstractMetaFunctionCList>; // Sorted
 
     ShibokenGenerator();
@@ -97,6 +99,10 @@ public:
     static QString minimalConstructorExpression(const ApiExtractorResult &api,
                                                 const TypeEntryCPtr &type);
 
+    /// Return the name of the _TypeF() function generated to get the PyTypeObject
+    static QString cpythonTypeName(const AbstractMetaClassCPtr &metaClass);
+    static QString cpythonTypeName(const ComplexTypeEntryCPtr &type);
+
 protected:
     bool doSetup() override;
 
@@ -109,6 +115,15 @@ protected:
      */
     FunctionGroups getGlobalFunctionGroups() const;
     static FunctionGroups getFunctionGroups(const AbstractMetaClassCPtr &scope);
+    /// Returns the constructors for which bindings should be generated.
+    /// \param scope Where to search for functions
+    static AbstractMetaFunctionCList getConstructors(const AbstractMetaClassCPtr &scope);
+    /// Returns the constructors which should be added to wrapper class.
+    /// \param scope Where to search for functions
+    static AbstractMetaFunctionCList getWrapperConstructors(const AbstractMetaClassCPtr &scope);
+    /// Returns mapping of virtual functions whose override handling code can be re-used.
+    static const FunctionMapping &
+        getReusedOverridenFunctions(const AbstractMetaClassCPtr &scope);
 
     static QList<AbstractMetaFunctionCList>
         numberProtocolOperators(const AbstractMetaClassCPtr &scope);
@@ -147,6 +162,8 @@ protected:
     void processCodeSnip(QString &code) const;
     void processCodeSnip(QString &code, const QString &context) const;
     void processClassCodeSnip(QString &code, const GeneratorContext &context) const;
+    /// Replaces variables in a custom conversion type check snippet
+    void processTypeCheckCodeSnip(QString &code, const QString &context) const;
 
     /**
      *   Verifies if any of the function's code injections makes a call
@@ -165,8 +182,8 @@ protected:
      *   \param arg_count the number of function arguments
      */
     QString functionSignature(const AbstractMetaFunctionCPtr &func,
-                              const QString &prepend = QString(),
-                              const QString &append = QString(),
+                              const QString &className = {},
+                              const QString &append = {},
                               Options options = NoOption,
                               int arg_count = -1) const;
 
@@ -215,10 +232,12 @@ protected:
 
     static QString fullPythonClassName(const AbstractMetaClassCPtr &metaClass);
 
+    static QString headerFileNameForClass(const AbstractMetaClassCPtr &metaClass);
     static QString headerFileNameForContext(const GeneratorContext &context);
     IncludeGroup baseWrapperIncludes(const GeneratorContext &classContext) const;
 
     static QString fullPythonFunctionName(const AbstractMetaFunctionCPtr &func, bool forceFunc);
+    static QString pythonOverrideImplName(const AbstractMetaFunctionCPtr &func);
 
     static bool wrapperDiagnostics();
 
@@ -246,12 +265,10 @@ protected:
     static QString converterObject(const AbstractMetaType &type) ;
     static QString converterObject(const TypeEntryCPtr &type);
 
+    /// Return a name prefixed by Sbk_ which can be used for naming variables in the code
     static QString cpythonBaseName(const AbstractMetaClassCPtr &metaClass);
-    static QString cpythonBaseName(const TypeEntryCPtr &type);
+    static QString cpythonBaseName(const ComplexTypeEntryCPtr &type);
     static QString containerCpythonBaseName(const ContainerTypeEntryCPtr &ctype);
-    static QString cpythonBaseName(const AbstractMetaType &type);
-    static QString cpythonTypeName(const AbstractMetaClassCPtr &metaClass);
-    static QString cpythonTypeName(const TypeEntryCPtr &type);
     static QString cpythonTypeNameExtSet(const TypeEntryCPtr &type);
     static QString cpythonTypeNameExtSet(const AbstractMetaType &type);
     static QString cpythonTypeNameExt(const TypeEntryCPtr &type);
@@ -268,6 +285,7 @@ protected:
     static QString cpythonToPythonConversionFunction(const AbstractMetaClassCPtr &metaClass);
     static QString cpythonToPythonConversionFunction(const TypeEntryCPtr &type);
 
+    static QString cpythonConstructorName(const AbstractMetaClassCPtr &metaClass);
     static QString cpythonFunctionName(const AbstractMetaFunctionCPtr &func) ;
     static QString cpythonMethodDefinitionName(const AbstractMetaFunctionCPtr &func);
     static QString cpythonGettersSettersDefinitionName(const AbstractMetaClassCPtr &metaClass);
@@ -369,11 +387,18 @@ private:
 
     static const GeneratorClassInfoCacheEntry &
         getGeneratorClassInfo(const AbstractMetaClassCPtr &scope);
-    static FunctionGroups getFunctionGroupsImpl(const AbstractMetaClassCPtr &scope);
+    static FunctionGroups getFunctionGroupsImpl(const AbstractMetaClassCPtr &scope,
+                                                AbstractMetaFunctionCList *constructors);
+    static AbstractMetaFunctionCList wrapperConstructorsImpl(const AbstractMetaClassCPtr &scope);
+    static AbstractMetaFunctionCList getWrapperConstructorsFromUsingDirective(const AbstractMetaClassCPtr &scope);
     static QList<AbstractMetaFunctionCList>
         getNumberProtocolOperators(const AbstractMetaClassCPtr &metaClass);
     static BoolCastFunctionOptional getBoolCast(const AbstractMetaClassCPtr &metaClass);
-    static bool classNeedsGetattroFunctionImpl(const AbstractMetaClassCPtr &metaClass);
+    static bool classNeedsGetattroOverloadFunctionImpl(const FunctionGroups &functionGroups);
+    static AttroCheck checkAttroFunctionNeedsImpl(const AbstractMetaClassCPtr &metaClass,
+                                                  const FunctionGroups &functionGroups);
+    static FunctionMapping getReusedOverridesImpl(const AbstractMetaClassCPtr &metaClass);
+    static bool isVirtualOverride(const AbstractMetaFunctionCPtr &f);
 
     QString translateTypeForWrapperMethod(const AbstractMetaType &cType,
                                           const AbstractMetaClassCPtr &context,
@@ -437,22 +462,22 @@ private:
                                             QString &code) const;
 
     /// Replaces the %CONVERTTOPYTHON type system variable.
-    inline void replaceConvertToPythonTypeSystemVariable(QString &code) const
+    void replaceConvertToPythonTypeSystemVariable(QString &code) const
     {
         replaceConverterTypeSystemVariable(TypeSystemToPythonFunction, code);
     }
     /// Replaces the %CONVERTTOCPP type system variable.
-    inline void replaceConvertToCppTypeSystemVariable(QString &code) const
+    void replaceConvertToCppTypeSystemVariable(QString &code) const
     {
         replaceConverterTypeSystemVariable(TypeSystemToCppFunction, code);
     }
     /// Replaces the %ISCONVERTIBLE type system variable.
-    inline void replaceIsConvertibleToCppTypeSystemVariable(QString &code) const
+    void replaceIsConvertibleToCppTypeSystemVariable(QString &code) const
     {
         replaceConverterTypeSystemVariable(TypeSystemIsConvertibleFunction, code);
     }
     /// Replaces the %CHECKTYPE type system variable.
-    inline void replaceTypeCheckTypeSystemVariable(QString &code) const
+    void replaceTypeCheckTypeSystemVariable(QString &code) const
     {
         replaceConverterTypeSystemVariable(TypeSystemCheckFunction, code);
     }
