@@ -1,15 +1,15 @@
-// Copyright (C) 2018 The Qt Company Ltd.
+// Copyright (C) 2023 The Qt Company Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
-#include "sbkpython.h"
+#define PEP384_INTERN
+
+#include "pep384impl.h"
 #include "autodecref.h"
 #include "sbkstaticstrings.h"
 #include "sbkstaticstrings_p.h"
 #include "basewrapper.h"
 #include "basewrapper_p.h"
 #include "sbkenum.h"
-#include "sbkenum_p.h"
-#include "sbkconverter.h"
 #include "voidptr.h"
 
 #include <cstdlib>
@@ -105,20 +105,22 @@ static PyType_Spec typeprobe_spec = {
 static void
 check_PyTypeObject_valid()
 {
-    auto *obtype = reinterpret_cast<PyObject *>(&PyType_Type);
-    auto *probe_tp_base = reinterpret_cast<PyTypeObject *>(
-        PyObject_GetAttr(obtype, Shiboken::PyMagicName::base()));
+    auto *typetype = &PyType_Type;
+    auto *obtype = reinterpret_cast<PyObject *>(typetype);
+    auto *probe_tp_base_obj = PyObject_GetAttr(obtype, Shiboken::PyMagicName::base());
+    auto *probe_tp_base = reinterpret_cast<PyTypeObject *>(probe_tp_base_obj);
     auto *probe_tp_bases = PyObject_GetAttr(obtype, Shiboken::PyMagicName::bases());
-    auto *check = reinterpret_cast<PyTypeObject *>(
-        PyType_FromSpecWithBases(&typeprobe_spec, probe_tp_bases));
-    auto *typetype = reinterpret_cast<PyTypeObject *>(obtype);
+    auto *checkObj = PyType_FromSpecWithBases(&typeprobe_spec, probe_tp_bases);
+    auto *check = reinterpret_cast<PyTypeObject *>(checkObj);
     PyObject *w = PyObject_GetAttr(obtype, Shiboken::PyMagicName::weakrefoffset());
     long probe_tp_weakrefoffset = PyLong_AsLong(w);
     PyObject *d = PyObject_GetAttr(obtype, Shiboken::PyMagicName::dictoffset());
     long probe_tp_dictoffset = PyLong_AsLong(d);
     PyObject *probe_tp_mro = PyObject_GetAttr(obtype, Shiboken::PyMagicName::mro());
+    Shiboken::AutoDecRef tpDict(PepType_GetDict(check));
+    auto *checkDict = tpDict.object();
     if (false
-        || strcmp(probe_tp_name, check->tp_name) != 0
+        || std::strcmp(probe_tp_name, check->tp_name) != 0
         || probe_tp_basicsize       != check->tp_basicsize
         || probe_tp_dealloc         != check->tp_dealloc
         || probe_tp_repr            != check->tp_repr
@@ -133,8 +135,8 @@ check_PyTypeObject_valid()
         || probe_tp_methods         != check->tp_methods
         || probe_tp_getset          != check->tp_getset
         || probe_tp_base            != typetype->tp_base
-        || !PyDict_Check(check->tp_dict)
-        || !PyDict_GetItemString(check->tp_dict, "dummy")
+        || !PyDict_Check(checkDict)
+        || !PyDict_GetItemString(checkDict, "dummy")
         || probe_tp_descr_get       != check->tp_descr_get
         || probe_tp_descr_set       != check->tp_descr_set
         || probe_tp_dictoffset      != typetype->tp_dictoffset
@@ -146,18 +148,15 @@ check_PyTypeObject_valid()
         || probe_tp_bases           != typetype->tp_bases
         || probe_tp_mro             != typetype->tp_mro
         || Py_TPFLAGS_DEFAULT       != (check->tp_flags & Py_TPFLAGS_DEFAULT))
-        Py_FatalError("The structure of type objects has changed!");
-    Py_DECREF(check);
-    Py_DECREF(probe_tp_base);
+        Py_FatalError("libshiboken: The structure of type objects has changed!");
+    Py_DECREF(checkObj);
+    Py_DECREF(probe_tp_base_obj);
     Py_DECREF(w);
     Py_DECREF(d);
     Py_DECREF(probe_tp_bases);
     Py_DECREF(probe_tp_mro);
 }
 
-// PYSIDE-1797: This must be a runtime decision.
-#include "pep384_issue33738.cpp"
-
 #endif // Py_LIMITED_API
 
 /*****************************************************************************
@@ -178,35 +177,27 @@ check_PyTypeObject_valid()
 static PyObject *
 find_name_in_mro(PyTypeObject *type, PyObject *name, int *error)
 {
-    Py_ssize_t i, n;
-    PyObject *mro, *res, *base, *dict;
-
-    /* Look in tp_dict of types in MRO */
-    mro = type->tp_mro;
-
-    res = nullptr;
-    /* Keep a strong reference to mro because type->tp_mro can be replaced
-       during dict lookup, e.g. when comparing to non-string keys. */
-    Py_INCREF(mro);
-    assert(PyTuple_Check(mro));
-    n = PyTuple_GET_SIZE(mro);
-    for (i = 0; i < n; i++) {
-        base = PyTuple_GET_ITEM(mro, i);
+    *error = 0;
+    /* Look in tp_dict of types in MRO. Keep a strong reference to mro because
+     * type->tp_mro can be replaced during dict lookup, e.g. when comparing to
+     * non-string keys. */
+    assert(PyTuple_Check(type->tp_mro));
+    Py_INCREF(type->tp_mro);
+    Shiboken::AutoDecRef mro(type->tp_mro);
+    for (Py_ssize_t i = 0, n = PyTuple_Size(mro.object()); i < n; ++i) {
+        auto *base = PyTuple_GetItem(mro.object(), i);
         assert(PyType_Check(base));
-        dict = ((PyTypeObject *)base)->tp_dict;
-        assert(dict && PyDict_Check(dict));
-        res = PyDict_GetItem(dict, name);
-        if (res != nullptr)
-            break;
+        auto *type = reinterpret_cast<PyTypeObject *>(base);
+        Shiboken::AutoDecRef dict(PepType_GetDict(type));
+        assert(!dict.isNull() && PyDict_Check(dict.object()));
+        if (auto *res = PyDict_GetItem(dict.object(), name))
+            return res;
         if (PyErr_Occurred()) {
             *error = -1;
-            goto done;
+            return nullptr;
         }
     }
-    *error = 0;
-done:
-    Py_DECREF(mro);
-    return res;
+    return nullptr;
 }
 
 /* Internal API to look for a name through the MRO.
@@ -214,26 +205,20 @@ done:
 PyObject *
 _PepType_Lookup(PyTypeObject *type, PyObject *name)
 {
-    PyObject *res;
-    int error;
-
     /* We may end up clearing live exceptions below, so make sure it's ours. */
     assert(!PyErr_Occurred());
 
-    res = find_name_in_mro(type, name, &error);
-    /* Only put NULL results into cache if there was no error. */
+    int error{};
+    auto *res = find_name_in_mro(type, name, &error);
+    /* Only put nullptr results into cache if there was no error. */
     if (error) {
-        /* It's not ideal to clear the error condition,
-           but this function is documented as not setting
-           an exception, and I don't want to change that.
-           E.g., when PyType_Ready() can't proceed, it won't
-           set the "ready" flag, so future attempts to ready
-           the same type will call it again -- hopefully
-           in a context that propagates the exception out.
-        */
-        if (error == -1) {
+        /* It's not ideal to clear the error condition, but this function is
+         * documented as not setting an exception, and I don't want to change
+         * that. E.g., when PyType_Ready() can't proceed, it won't set the
+         * "ready" flag, so future attempts to ready the same type will call
+         * it again -- hopefully in a context that propagates the exception out. */
+        if (error == -1)
             PyErr_Clear();
-        }
         return nullptr;
     }
     return res;
@@ -251,7 +236,7 @@ _PepType_Lookup(PyTypeObject *type, PyObject *name)
 // structs and macros modelled after their equivalents in
 // cpython/Include/cpython/unicodeobject.h
 
-struct PepASCIIObject
+struct PepASCIIObject // since 3.12
 {
     PyObject_HEAD
     Py_ssize_t length;          /* Number of code points in the string */
@@ -264,18 +249,29 @@ struct PepASCIIObject
         unsigned int ready:1;
         unsigned int :24;
     } state;
+};
+
+struct PepASCIIObject_311 : public PepASCIIObject
+{
     wchar_t *wstr;              /* wchar_t representation (null-terminated) */
 };
 
-struct PepCompactUnicodeObject
+struct PepCompactUnicodeObject // since 3.12
 {
     PepASCIIObject _base;
     Py_ssize_t utf8_length;
     char *utf8;                 /* UTF-8 representation (null-terminated) */
+};
+
+struct PepCompactUnicodeObject_311 // since 3.12
+{
+    PepASCIIObject_311 _base;
+    Py_ssize_t utf8_length;
+    char *utf8;                 /* UTF-8 representation (null-terminated) */
     Py_ssize_t wstr_length;     /* Number of code points in wstr */
 };
 
-struct PepUnicodeObject
+struct PepUnicodeObject // since 3.12
 {
     PepCompactUnicodeObject _base;
     union {
@@ -286,6 +282,17 @@ struct PepUnicodeObject
     } data;                     /* Canonical, smallest-form Unicode buffer */
 };
 
+struct PepUnicodeObject_311
+{
+    PepCompactUnicodeObject_311 _base;
+    union {
+        void *any;
+        Py_UCS1 *latin1;
+        Py_UCS2 *ucs2;
+        Py_UCS4 *ucs4;
+    } data;                     /* Canonical, smallest-form Unicode buffer */
+};
+
 int _PepUnicode_KIND(PyObject *str)
 {
     return reinterpret_cast<PepASCIIObject *>(str)->state.kind;
@@ -303,18 +310,33 @@ int _PepUnicode_IS_COMPACT(PyObject *str)
     return asciiObj->state.compact;
 }
 
+static void *_PepUnicode_ASCII_DATA(PyObject *str)
+{
+    if (_PepRuntimeVersion() < 0x030C00) {
+        auto *asciiObj_311 = reinterpret_cast<PepASCIIObject_311 *>(str);
+        return asciiObj_311 + 1;
+    }
+    auto *asciiObj = reinterpret_cast<PepASCIIObject *>(str);
+    return asciiObj + 1;
+}
+
 static void *_PepUnicode_COMPACT_DATA(PyObject *str)
 {
-    auto *asciiObj = reinterpret_cast<PepASCIIObject *>(str);
-    if (asciiObj->state.ascii)
-        return asciiObj + 1;
+    if (_PepUnicode_IS_ASCII(str) != 0)
+        return _PepUnicode_ASCII_DATA(str);
+    if (_PepRuntimeVersion() < 0x030C00) {
+        auto *compactObj_311 = reinterpret_cast<PepCompactUnicodeObject_311 *>(str);
+        return compactObj_311 + 1;
+    }
     auto *compactObj = reinterpret_cast<PepCompactUnicodeObject *>(str);
     return compactObj + 1;
 }
 
 static void *_PepUnicode_NONCOMPACT_DATA(PyObject *str)
 {
-    return reinterpret_cast<PepUnicodeObject *>(str)->data.any;
+    return _PepRuntimeVersion() < 0x030C00
+        ? reinterpret_cast<PepUnicodeObject_311 *>(str)->data.any
+        : reinterpret_cast<PepUnicodeObject *>(str)->data.any;
 }
 
 void *_PepUnicode_DATA(PyObject *str)
@@ -325,6 +347,23 @@ void *_PepUnicode_DATA(PyObject *str)
 
 // Fast path accessing UTF8 data without doing a conversion similar
 // to _PyUnicode_AsUTF8String
+static const char *utf8FastPath_311(PyObject *str)
+{
+    if (PyUnicode_GetLength(str) == 0)
+        return "";
+    auto *asciiObj = reinterpret_cast<PepASCIIObject_311 *>(str);
+    if (asciiObj->state.kind != PepUnicode_1BYTE_KIND || asciiObj->state.compact == 0)
+        return nullptr; // Empirical: PyCompactUnicodeObject.utf8 is only valid for 1 byte
+    if (asciiObj->state.ascii) {
+        auto *data = asciiObj + 1;
+        return reinterpret_cast<const char *>(data);
+    }
+    auto *compactObj = reinterpret_cast<PepCompactUnicodeObject_311 *>(str);
+    if (compactObj->utf8_length)
+        return compactObj->utf8;
+    return nullptr;
+}
+
 static const char *utf8FastPath(PyObject *str)
 {
     if (PyUnicode_GetLength(str) == 0)
@@ -356,8 +395,10 @@ const char *_PepUnicode_AsString(PyObject *str)
 #define TOSTRING(x) STRINGIFY(x)
 #define AT __FILE__ ":" TOSTRING(__LINE__)
 
-    if (const auto *utf8 = utf8FastPath(str))
+    if (const auto *utf8 = _PepRuntimeVersion() < 0x030C00
+        ? utf8FastPath_311(str) : utf8FastPath(str)) {
         return utf8;
+    }
 
     static PyObject *cstring_dict = nullptr;
     if (cstring_dict == nullptr) {
@@ -366,6 +407,8 @@ const char *_PepUnicode_AsString(PyObject *str)
             Py_FatalError("Error in " AT);
     }
     PyObject *bytesStr = PyUnicode_AsEncodedString(str, "utf8", nullptr);
+    if (bytesStr == nullptr)
+        Py_FatalError("Error in " AT);
     PyObject *entry = PyDict_GetItemWithError(cstring_dict, bytesStr);
     if (entry == nullptr) {
         int e = PyDict_SetItem(cstring_dict, bytesStr, bytesStr);
@@ -386,13 +429,11 @@ const char *_PepUnicode_AsString(PyObject *str)
  */
 #ifdef Py_LIMITED_API
 
-static PyObject *sys_flags = nullptr;
-
 int
 Pep_GetFlag(const char *name)
 {
+    static PyObject *sys_flags = nullptr;
     static int initialized = 0;
-    int ret = -1;
 
     if (!initialized) {
         sys_flags = PySys_GetObject("flags");
@@ -401,14 +442,11 @@ Pep_GetFlag(const char *name)
         initialized = 1;
     }
     if (sys_flags != nullptr) {
-        PyObject *ob_ret = PyObject_GetAttrString(sys_flags, name);
-        if (ob_ret != nullptr) {
-            long long_ret = PyLong_AsLong(ob_ret);
-            Py_DECREF(ob_ret);
-            ret = (int) long_ret;
-        }
+        Shiboken::AutoDecRef ob_ret(PyObject_GetAttrString(sys_flags, name));
+        if (!ob_ret.isNull())
+            return int(PyLong_AsLong(ob_ret.object()));
     }
-    return ret;
+    return -1;
 }
 
 int
@@ -426,6 +464,47 @@ Pep_GetVerboseFlag()
 }
 #endif // Py_LIMITED_API
 
+// Support for pyerrors.h
+
+#ifdef PEP_OLD_ERR_API
+// Emulate PyErr_GetRaisedException() using the deprecated PyErr_Fetch()/PyErr_Store()
+PyObject *PepErr_GetRaisedException()
+{
+    PyObject *type{};
+    PyObject *value{};
+    PyObject *traceback{};
+    PyErr_Fetch(&type, &value, &traceback);
+    Py_XINCREF(value);
+    PyErr_Restore(type, value, traceback);
+    return value;
+}
+
+struct PepException_HEAD
+{
+    PyObject_HEAD
+    PyObject *x1; // dict
+    PyObject *args;
+};
+
+// PyException_GetArgs/PyException_SetArgs were added to the stable API in 3.12
+PyObject *PepException_GetArgs(PyObject *ex)
+{
+    auto *h = reinterpret_cast<PepException_HEAD *>(ex);
+    Py_XINCREF(h->args);
+    return h->args;
+}
+
+LIBSHIBOKEN_API void PepException_SetArgs(PyObject *ex, PyObject *args)
+{
+    auto *h = reinterpret_cast<PepException_HEAD *>(ex);
+    Py_XINCREF(args);
+    auto *old = h->args; // Py_XSETREF()
+    h->args = args;
+    Py_XDECREF(old);
+
+}
+#endif // Limited or < 3.12
+
 /*****************************************************************************
  *
  * Support for code.h
@@ -436,20 +515,28 @@ Pep_GetVerboseFlag()
 int
 PepCode_Get(PepCodeObject *co, const char *name)
 {
-    PyObject *ob = reinterpret_cast<PyObject *>(co);
-    PyObject *ob_ret;
-    int ret = -1;
-
-    ob_ret = PyObject_GetAttrString(ob, name);
-    if (ob_ret != nullptr) {
-        long long_ret = PyLong_AsLong(ob_ret);
-        Py_DECREF(ob_ret);
-        ret = (int) long_ret;
-    }
-    return ret;
+    auto *ob = reinterpret_cast<PyObject *>(co);
+    Shiboken::AutoDecRef ob_ret(PyObject_GetAttrString(ob, name));
+    return ob_ret.isNull() ? -1 : int(PyLong_AsLong(ob_ret.object()));
 }
+
+int PepCode_Check(PyObject *o)
+{
+    return o != nullptr && std::strcmp(Py_TYPE(o)->tp_name, "code") == 0 ? 1 : 0;
+}
+
 #endif // Py_LIMITED_API
 
+#if defined(Py_LIMITED_API) || defined(PYPY_VERSION)
+PyObject *PepFunction_GetDefaults(PyObject *function)
+{
+    auto *ob_ret = PyObject_GetAttrString(function, "__defaults__");
+    Py_XDECREF(ob_ret); // returns borrowed ref
+    return ob_ret != Py_None ? ob_ret : nullptr;
+}
+
+#endif // defined(Py_LIMITED_API) || defined(PYPY_VERSION)
+
 /*****************************************************************************
  *
  * Support for datetime.h
@@ -464,7 +551,7 @@ static PyTypeObject *dt_getCheck(const char *name)
     PyObject *op = PyObject_GetAttrString(PyDateTimeAPI->module, name);
     if (op == nullptr) {
         fprintf(stderr, "datetime.%s not found\n", name);
-        Py_FatalError("aborting");
+        Py_FatalError("libshiboken: error initializing DateTime support, aborting");
     }
     return reinterpret_cast<PyTypeObject *>(op);
 }
@@ -478,10 +565,10 @@ init_DateTime(void)
     if (!initialized) {
         PyDateTimeAPI = (datetime_struc *)malloc(sizeof(datetime_struc));
         if (PyDateTimeAPI == nullptr)
-            Py_FatalError("PyDateTimeAPI malloc error, aborting");
+            Py_FatalError("libshiboken: PyDateTimeAPI malloc error, aborting");
         PyDateTimeAPI->module = PyImport_ImportModule("datetime");
         if (PyDateTimeAPI->module == nullptr)
-            Py_FatalError("datetime module not found, aborting");
+            Py_FatalError("libshiboken: datetime module not found, aborting");
         PyDateTimeAPI->DateType     = dt_getCheck("date");
         PyDateTimeAPI->DateTimeType = dt_getCheck("datetime");
         PyDateTimeAPI->TimeType     = dt_getCheck("time");
@@ -495,32 +582,24 @@ init_DateTime(void)
 int
 PyDateTime_Get(PyObject *ob, const char *name)
 {
-    PyObject *ob_ret;
-    int ret = -1;
-
-    ob_ret = PyObject_GetAttrString(ob, name);
-    if (ob_ret != nullptr) {
-        long long_ret = PyLong_AsLong(ob_ret);
-        Py_DECREF(ob_ret);
-        ret = (int) long_ret;
-    }
-    return ret;
+    Shiboken::AutoDecRef ob_ret(PyObject_GetAttrString(ob, name));
+    return ob_ret.isNull() ? -1 : int(PyLong_AsLong(ob_ret.object()));
 }
 
 PyObject *
 PyDate_FromDate(int year, int month, int day)
 {
-    return PyObject_CallFunction((PyObject *)PyDateTimeAPI->DateType,
-                                 (char *)"(iii)", year, month, day);
+    auto *ob = reinterpret_cast<PyObject *>(PyDateTimeAPI->DateType);
+    return PyObject_CallFunction(ob, "(iii)", year, month, day);
 }
 
 PyObject *
 PyDateTime_FromDateAndTime(int year, int month, int day,
                            int hour, int min, int sec, int usec)
 {
-    return PyObject_CallFunction((PyObject *)PyDateTimeAPI->DateTimeType,
-                                 (char *)"(iiiiiii)", year, month, day,
-                                  hour, min, sec, usec);
+    auto *ob = reinterpret_cast<PyObject *>(PyDateTimeAPI->DateTimeType);
+    return PyObject_CallFunction(ob, "(iiiiiii)", year, month, day,
+                                 hour, min, sec, usec);
 }
 
 PyObject *
@@ -542,14 +621,8 @@ PyTime_FromTime(int hour, int min, int sec, int usec)
 PyObject *
 PyRun_String(const char *str, int start, PyObject *globals, PyObject *locals)
 {
-    PyObject *code = Py_CompileString(str, "pyscript", start);
-    PyObject *ret = nullptr;
-
-    if (code != nullptr) {
-        ret = PyEval_EvalCode(code, globals, locals);
-    }
-    Py_XDECREF(code);
-    return ret;
+    Shiboken::AutoDecRef code(Py_CompileString(str, "pyscript", start));
+    return code.isNull() ? nullptr : PyEval_EvalCode(code.object(), globals, locals);
 }
 
 #endif // Py_LIMITED_API
@@ -565,10 +638,10 @@ PyTypeObject *PepMethod_TypePtr = nullptr;
 
 static PyTypeObject *getMethodType(void)
 {
-    static const char prog[] =
-        "class _C:\n"
-        "    def _m(self): pass\n"
-        "result = type(_C()._m)\n";
+    static const char prog[] = R"(class _C:
+    def _m(self): pass
+result = type(_C()._m)
+)";
     return reinterpret_cast<PyTypeObject *>(PepRun_GetResult(prog));
 }
 
@@ -576,16 +649,15 @@ static PyTypeObject *getMethodType(void)
 PyObject *
 PyMethod_New(PyObject *func, PyObject *self)
 {
-    return PyObject_CallFunction((PyObject *)PepMethod_TypePtr,
-                                 (char *)"(OO)", func, self);
+    auto *ob = reinterpret_cast<PyObject *>(PepMethod_TypePtr);
+    return PyObject_CallFunction(ob, "(OO)", func, self);
 }
 
 PyObject *
 PyMethod_Function(PyObject *im)
 {
-    PyObject *ret = PyObject_GetAttr(im, Shiboken::PyMagicName::func());
-
     // We have to return a borrowed reference.
+    PyObject *ret = PyObject_GetAttr(im, Shiboken::PyMagicName::func());
     Py_DECREF(ret);
     return ret;
 }
@@ -593,10 +665,9 @@ PyMethod_Function(PyObject *im)
 PyObject *
 PyMethod_Self(PyObject *im)
 {
-    PyObject *ret = PyObject_GetAttr(im, Shiboken::PyMagicName::self());
-
     // We have to return a borrowed reference.
     // If we don't obey that here, then we get a test error!
+    PyObject *ret = PyObject_GetAttr(im, Shiboken::PyMagicName::self());
     Py_DECREF(ret);
     return ret;
 }
@@ -612,10 +683,8 @@ PyMethod_Self(PyObject *im)
 PyObject *
 PepFunction_Get(PyObject *ob, const char *name)
 {
-    PyObject *ret;
-
     // We have to return a borrowed reference.
-    ret = PyObject_GetAttrString(ob, name);
+    PyObject *ret = PyObject_GetAttrString(ob, name);
     Py_XDECREF(ret);
     return ret;
 }
@@ -630,7 +699,7 @@ static PyTypeObject *getFunctionType(void)
         "from types import FunctionType as result\n";
     return reinterpret_cast<PyTypeObject *>(PepRun_GetResult(prog));
 }
-#endif // Py_LIMITED_API || Python 2
+#endif // Py_LIMITED_API
 
 /*****************************************************************************
  *
@@ -650,11 +719,8 @@ PyTypeObject *PepStaticMethod_TypePtr = nullptr;
 static PyTypeObject *
 getStaticMethodType(void)
 {
-    // this works for Python 3, only
-    //    "StaticMethodType = type(str.__dict__['maketrans'])\n";
     static const char prog[] =
-        "from xxsubtype import spamlist\n"
-        "result = type(spamlist.__dict__['staticmeth'])\n";
+        "result = type(str.__dict__['maketrans'])\n";
     return reinterpret_cast<PyTypeObject *>(PepRun_GetResult(prog));
 }
 
@@ -714,9 +780,23 @@ PepType_GetNameStr(PyTypeObject *type)
 {
     const char *ret = type->tp_name;
     const char *nodots = std::strrchr(ret, '.');
-    if (nodots)
-        ret = nodots + 1;
-    return ret;
+    return nodots != nullptr ? nodots + 1 : ret;
+}
+
+const char *PepType_GetFullyQualifiedNameStr(PyTypeObject *type)
+{
+#if 0
+    // Should look like the below code for Limited API >= 3.13, however, it crashes for heap types
+    //  since PyType_GetFullyQualifiedName() calls type_qualname() at Objects/typeobject.c:1402
+    // which expects a PyHeapTypeObject (struct _heaptypeobject).
+    Shiboken::AutoDecRef name(PyType_GetFullyQualifiedName(type));
+    Py_ssize_t size{};
+    const char *result = PyUnicode_AsUTF8AndSize(name.object(), &size);
+    assert(result != nullptr && size > 0);
+    return result;
+#else
+    return type->tp_name;
+#endif
 }
 
 // PYSIDE-2264: Find the _functools or functools module and retrieve the
@@ -736,10 +816,10 @@ Pep_GetPartialFunction(void)
         functools = PyImport_ImportModule("functools");
     }
     if (!functools)
-        Py_FatalError("functools cannot be found");
+        Py_FatalError("libshiboken: functools cannot be found");
     result = PyObject_GetAttrString(functools, "partial");
     if (!result || !PyCallable_Check(result))
-        Py_FatalError("partial not found or not a function");
+        Py_FatalError("libshiboken: partial not found or not a function");
     initialized = true;
     return result;
 }
@@ -749,33 +829,6 @@ Pep_GetPartialFunction(void)
  * Newly introduced convenience functions
  *
  */
-#if PY_VERSION_HEX < 0x03070000 || defined(Py_LIMITED_API)
-
-PyObject *
-PyImport_GetModule(PyObject *name)
-{
-    PyObject *m;
-    PyObject *modules = PyImport_GetModuleDict();
-    if (modules == NULL) {
-        PyErr_SetString(PyExc_RuntimeError, "unable to get sys.modules");
-        return NULL;
-    }
-    Py_INCREF(modules);
-    if (PyDict_CheckExact(modules)) {
-        m = PyDict_GetItemWithError(modules, name);  /* borrowed */
-        Py_XINCREF(m);
-    }
-    else {
-        m = PyObject_GetItem(modules, name);
-        if (m == NULL && PyErr_ExceptionMatches(PyExc_KeyError)) {
-            PyErr_Clear();
-        }
-    }
-    Py_DECREF(modules);
-    return m;
-}
-
-#endif // PY_VERSION_HEX < 0x03070000 || defined(Py_LIMITED_API)
 
 // 2020-06-16: For simplicity of creating arbitrary things, this function
 // is now made public.
@@ -786,15 +839,17 @@ PepRun_GetResult(const char *command)
     /*
      * Evaluate a string and return the variable `result`
      */
-    PyObject *d, *v, *res;
-
-    d = PyDict_New();
-    if (d == nullptr
-        || PyDict_SetItem(d, Shiboken::PyMagicName::builtins(), PyEval_GetBuiltins()) < 0) {
+    PyObject *d = PyDict_New();
+    if (d == nullptr)
         return nullptr;
-    }
-    v = PyRun_String(command, Py_file_input, d, d);
-    res = v ? PyDict_GetItem(d, Shiboken::PyName::result()) : nullptr;
+
+    Shiboken::AutoDecRef builtins(PepEval_GetFrameBuiltins());
+    if (PyDict_SetItem(d, Shiboken::PyMagicName::builtins(), PyEval_GetBuiltins()) < 0)
+        return nullptr;
+    builtins.reset(nullptr);
+
+    PyObject *v = PyRun_String(command, Py_file_input, d, d);
+    PyObject *res = v ? PyDict_GetItem(d, Shiboken::PyName::result()) : nullptr;
     Py_XDECREF(v);
     Py_DECREF(d);
     return res;
@@ -802,7 +857,7 @@ PepRun_GetResult(const char *command)
 
 PyTypeObject *PepType_Type_tp_new(PyTypeObject *metatype, PyObject *args, PyObject *kwds)
 {
-    auto ret = PyType_Type.tp_new(metatype, args, kwds);
+    auto *ret = PyType_Type.tp_new(metatype, args, kwds);
     return reinterpret_cast<PyTypeObject *>(ret);
 }
 
@@ -812,12 +867,6 @@ PyTypeObject *PepType_Type_tp_new(PyTypeObject *metatype, PyObject *args, PyObje
  *
  */
 
-#ifdef Py_LIMITED_API
-// We keep these definitions local, because they don't work in Python 2.
-# define PyUnicode_GET_LENGTH(op)    PyUnicode_GetLength((PyObject *)(op))
-# define PyUnicode_READ_CHAR(u, i)   PyUnicode_ReadChar((PyObject *)(u), (i))
-#endif // Py_LIMITED_API
-
 PyObject *
 _Pep_PrivateMangle(PyObject *self, PyObject *name)
 {
@@ -826,15 +875,15 @@ _Pep_PrivateMangle(PyObject *self, PyObject *name)
      * This function is modelled after _Py_Mangle, but is optimized
      * a little for our purpose.
      */
-    if (PyUnicode_READ_CHAR(name, 0) != '_' ||
-        PyUnicode_READ_CHAR(name, 1) != '_') {
+    if (PyUnicode_ReadChar(name, 0) != '_' ||
+        PyUnicode_ReadChar(name, 1) != '_') {
         Py_INCREF(name);
         return name;
     }
-    size_t nlen = PyUnicode_GET_LENGTH(name);
+    const Py_ssize_t nlen = PyUnicode_GetLength(name);
     /* Don't mangle __id__ or names with dots. */
-    if ((PyUnicode_READ_CHAR(name, nlen-1) == '_' &&
-         PyUnicode_READ_CHAR(name, nlen-2) == '_') ||
+    if ((PyUnicode_ReadChar(name, nlen-1) == '_' &&
+         PyUnicode_ReadChar(name, nlen-2) == '_') ||
         PyUnicode_FindChar(name, '.', 0, nlen, 1) != -1) {
         Py_INCREF(name);
         return name;
@@ -845,10 +894,10 @@ _Pep_PrivateMangle(PyObject *self, PyObject *name)
     // PYSIDE-1436: _Py_Mangle is no longer exposed; implement it always.
     // The rest of this function is our own implementation of _Py_Mangle.
     // Please compare the original function in compile.c .
-    size_t plen = PyUnicode_GET_LENGTH(privateobj.object());
+    Py_ssize_t plen = PyUnicode_GetLength(privateobj.object());
     /* Strip leading underscores from class name */
-    size_t ipriv = 0;
-    while (PyUnicode_READ_CHAR(privateobj.object(), ipriv) == '_')
+    Py_ssize_t ipriv = 0;
+    while (PyUnicode_ReadChar(privateobj.object(), ipriv) == '_')
         ipriv++;
     if (ipriv == plen) {
         Py_INCREF(name);
@@ -861,10 +910,11 @@ _Pep_PrivateMangle(PyObject *self, PyObject *name)
                         "private identifier too large to be mangled");
         return nullptr;
     }
-    size_t const amount = ipriv + 1 + plen + nlen;
-    size_t const big_stack = 1000;
+    const Py_ssize_t amount = ipriv + 1 + plen + nlen;
+    const Py_ssize_t big_stack = 1000;
     wchar_t bigbuf[big_stack];
-    wchar_t *resbuf = amount <= big_stack ? bigbuf : (wchar_t *)malloc(sizeof(wchar_t) * amount);
+    wchar_t *resbuf = amount <= big_stack
+        ? bigbuf : reinterpret_cast<wchar_t *>(malloc(sizeof(wchar_t) * amount));
     if (!resbuf)
         return nullptr;
     /* ident = "_" + priv[ipriv:] + ident # i.e. 1+plen+nlen bytes */
@@ -885,26 +935,18 @@ _Pep_PrivateMangle(PyObject *self, PyObject *name)
  *
  */
 
-int PepRuntime_38_flag = 0;
-
-static void
-init_PepRuntime()
+static long _GetPepRuntimeVersion()
 {
-    // We expect a string of the form "\d\.\d+\."
-    const char *version = Py_GetVersion();
-    if (version[0] < '3')
-        return;
-    if (std::atoi(version + 2) >= 8)
-        PepRuntime_38_flag = 1;
+    auto *version = PySys_GetObject("version_info");
+    const auto major = PyLong_AsLong(PyTuple_GetItem(version, 0));
+    const auto minor = PyLong_AsLong(PyTuple_GetItem(version, 1));
+    const auto micro = PyLong_AsLong(PyTuple_GetItem(version, 2));
+    return major << 16 | minor << 8 | micro;
 }
 
 long _PepRuntimeVersion()
 {
-    static auto *version = PySys_GetObject("version_info");
-    static auto major = PyLong_AsLong(PyTuple_GetItem(version, 0));
-    static auto minor = PyLong_AsLong(PyTuple_GetItem(version, 1));
-    static auto micro = PyLong_AsLong(PyTuple_GetItem(version, 2));
-    static auto number = major << 16 | minor << 8 | micro;
+    static const auto number = _GetPepRuntimeVersion();
     return number;
 }
 
@@ -917,33 +959,108 @@ long _PepRuntimeVersion()
  *
  */
 
+///////////////////////////////////////////////////////////////////////
+//
+// PEP 697: Support for embedded type structures.
+//
+// According to `https://docs.python.org/3/c-api/object.html?highlight=pyobject_gettypedata#c.PyObject_GetTypeData`
+// the function `PyObject_GetTypeData` should belong to the Stable API
+// since version 3.12.0, but it does not. We use instead some copies
+// from Python source code.
+
+#if !defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030C0000
+
+# define PepObject_GetTypeData PyObject_GetTypeData
+
+SbkObjectTypePrivate *PepType_SOTP(PyTypeObject *type)
+{
+    // PYSIDE-2676: Use the meta type explicitly.
+    //              A derived type would fail the offset calculation.
+    static auto *meta = SbkObjectType_TypeF();
+    assert(SbkObjectType_Check(type));
+    auto *obType = reinterpret_cast<PyObject *>(type);
+    void *data = PyObject_GetTypeData(obType, meta);
+    return reinterpret_cast<SbkObjectTypePrivate *>(data);
+}
+
+void PepType_SOTP_delete(PyTypeObject * /*type*/)
+{
+}
+
+#else
+
+// The following comments are directly copied from Python 3.12
+//
+
+// Make sure we have maximum alignment, even if the current compiler
+// does not support max_align_t. Note that:
+// - Autoconf reports alignment of unknown types to 0.
+// - 'long double' has maximum alignment on *most* platforms,
+//   looks like the best we can do for pre-C11 compilers.
+// - The value is tested, see test_alignof_max_align_t
+#  if !defined(ALIGNOF_MAX_ALIGN_T) || ALIGNOF_MAX_ALIGN_T == 0
+#    undef ALIGNOF_MAX_ALIGN_T
+#    define ALIGNOF_MAX_ALIGN_T alignof(long double)
+#  endif
+
+/* Align up to the nearest multiple of alignof(max_align_t)
+ * (like _Py_ALIGN_UP, but for a size rather than pointer)
+ */
+static Py_ssize_t _align_up(Py_ssize_t size)
+{
+    return (size + ALIGNOF_MAX_ALIGN_T - 1) & ~(ALIGNOF_MAX_ALIGN_T - 1);
+}
+
+static void *PepObject_GetTypeData(PyObject *obj, PyTypeObject *cls)
+{
+    assert(PyObject_TypeCheck(obj, cls));
+    return reinterpret_cast<char *>(obj) + _align_up(cls->tp_base->tp_basicsize);
+}
+//
+///////////////////////////////////////////////////////////////////////
+
 /*
  * PyTypeObject extender
  */
+
 static std::unordered_map<PyTypeObject *, SbkObjectTypePrivate > SOTP_extender{};
 static thread_local PyTypeObject *SOTP_key{};
 static thread_local SbkObjectTypePrivate *SOTP_value{};
 
-SbkObjectTypePrivate *PepType_SOTP(PyTypeObject *sbkType)
+SbkObjectTypePrivate *PepType_SOTP(PyTypeObject *type)
 {
-    if (sbkType == SOTP_key)
+    static auto *meta = SbkObjectType_TypeF();
+    static bool use_312 = _PepRuntimeVersion() >= 0x030C00;
+    assert(SbkObjectType_Check(type));
+    if (use_312) {
+        auto *obType = reinterpret_cast<PyObject *>(type);
+        void *data = PepObject_GetTypeData(obType, meta);
+        return reinterpret_cast<SbkObjectTypePrivate *>(data);
+    }
+    if (type == SOTP_key)
         return SOTP_value;
-    auto it = SOTP_extender.find(sbkType);
+    auto it = SOTP_extender.find(type);
     if (it == SOTP_extender.end()) {
-        it = SOTP_extender.insert({sbkType, {}}).first;
+        it = SOTP_extender.insert({type, {}}).first;
         memset(&it->second, 0, sizeof(SbkObjectTypePrivate));
     }
-    SOTP_key = sbkType;
+    SOTP_key = type;
     SOTP_value = &it->second;
     return SOTP_value;
 }
 
-void PepType_SOTP_delete(PyTypeObject *sbkType)
+void PepType_SOTP_delete(PyTypeObject *type)
 {
-    SOTP_extender.erase(sbkType);
+    static bool use_312 = _PepRuntimeVersion() >= 0x030C00;
+    assert(SbkObjectType_Check(type));
+    if (use_312)
+        return;
+    SOTP_extender.erase(type);
     SOTP_key = nullptr;
 }
 
+#endif // !defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030C0000
+
 /*
  * SbkEnumType extender
  */
@@ -953,13 +1070,12 @@ static thread_local SbkEnumTypePrivate *SETP_value{};
 
 SbkEnumTypePrivate *PepType_SETP(SbkEnumType *enumType)
 {
+    // PYSIDE-2230: This makes no sense at all for Enum types.
     if (enumType == SETP_key)
         return SETP_value;
     auto it = SETP_extender.find(enumType);
-    if (it == SETP_extender.end()) {
-        it = SETP_extender.insert({enumType, {}}).first;
-        memset(&it->second, 0, sizeof(SbkEnumTypePrivate));
-    }
+    if (it == SETP_extender.end())
+        it = SETP_extender.insert({enumType, SbkEnumTypePrivate{nullptr, nullptr}}).first;
     SETP_key = enumType;
     SETP_value = &it->second;
     return SETP_value;
@@ -971,38 +1087,131 @@ void PepType_SETP_delete(SbkEnumType *enumType)
     SETP_key = nullptr;
 }
 
-/*
- * PySideQFlagsType extender
- */
-static std::unordered_map<PySideQFlagsType *, SbkQFlagsTypePrivate> PFTP_extender{};
-static thread_local PySideQFlagsType *PFTP_key{};
-static thread_local SbkQFlagsTypePrivate *PFTP_value{};
-
-SbkQFlagsTypePrivate *PepType_PFTP(PySideQFlagsType *flagsType)
+#ifdef Py_LIMITED_API
+static PyObject *emulatePyType_GetDict(PyTypeObject *type)
 {
-    static PyTypeObject *enumMeta = getPyEnumMeta();
-    auto *mappedType = reinterpret_cast<PyTypeObject *>(flagsType);
-    auto *metaType = Py_TYPE(mappedType);
-    if (metaType == enumMeta) {
-        return reinterpret_cast<SbkQFlagsTypePrivate *>(
-            PepType_SETP(reinterpret_cast<SbkEnumType *>(flagsType)));
+    if (_PepRuntimeVersion() < 0x030C00 || type->tp_dict) {
+        auto *res = type->tp_dict;
+        Py_XINCREF(res);
+        return res;
     }
-    if (flagsType == PFTP_key)
-        return PFTP_value;
-    auto it = PFTP_extender.find(flagsType);
-    if (it == PFTP_extender.end()) {
-        it = PFTP_extender.insert({flagsType, {}}).first;
-        memset(&it->second, 0, sizeof(SbkQFlagsTypePrivate));
-    }
-    PFTP_key = flagsType;
-    PFTP_value = &it->second;
-    return PFTP_value;
+    // PYSIDE-2230: Here we are really cheating. We don't know how to
+    //              access an internal dict, and so we simply pretend
+    //              it were an empty dict. This works great for our types.
+    // This was an unexpectedly simple solution :D
+    return PyDict_New();
+}
+#endif
+
+// PyType_GetDict: replacement for <static type>.tp_dict, which is
+// zero for builtin types since 3.12.
+PyObject *PepType_GetDict(PyTypeObject *type)
+{
+#if !defined(Py_LIMITED_API)
+#  if PY_VERSION_HEX >= 0x030C0000
+    return PyType_GetDict(type);
+#  else
+    // pre 3.12 fallback code, mimicking the addref-behavior.
+    Py_XINCREF(type->tp_dict);
+    return type->tp_dict;
+#  endif
+#else
+    return emulatePyType_GetDict(type);
+#endif // Py_LIMITED_API
 }
 
-void PepType_PFTP_delete(PySideQFlagsType *flagsType)
+int PepType_SetDict(PyTypeObject *type, PyObject *dict)
 {
-    PFTP_extender.erase(flagsType);
-    PFTP_key = nullptr;
+    type->tp_dict = dict;
+    return 0;
+}
+
+// Pre 3.10, PyType_GetSlot() would only work for heap types.
+// FIXME: PyType_GetSlot() can be used unconditionally when the
+// minimum limited API version is >= 3.10.
+void *PepType_GetSlot(PyTypeObject *type, int aSlot)
+{
+    static const bool is310 = _PepRuntimeVersion() >= 0x030A00;
+    if (is310 || (type->tp_flags & Py_TPFLAGS_HEAPTYPE) != 0)
+        return PyType_GetSlot(type, aSlot);
+
+    switch (aSlot) {
+    case Py_tp_alloc:
+        return reinterpret_cast<void *>(type->tp_alloc);
+    case Py_tp_getattro:
+        return reinterpret_cast<void *>(type->tp_getattro);
+    case Py_tp_setattro:
+        return reinterpret_cast<void *>(type->tp_setattro);
+    case Py_tp_descr_get:
+        return reinterpret_cast<void *>(type->tp_descr_get);
+    case Py_tp_descr_set:
+        return reinterpret_cast<void *>(type->tp_descr_set);
+    case Py_tp_call:
+        return reinterpret_cast<void *>(type->tp_call);
+    case Py_tp_new:
+        return reinterpret_cast<void *>(type->tp_new);
+    case Py_tp_init:
+        return reinterpret_cast<void *>(type->tp_init);
+    case Py_tp_free:
+        return reinterpret_cast<void *>(type->tp_free);
+    }
+    assert(false);
+    return nullptr;
+}
+
+PyObject *PepEval_GetFrameGlobals()
+{
+    // PyEval_GetFrameGlobals() (added to stable ABI in 3.13) returns a new reference
+    // as opposed to deprecated PyEval_GetGlobals() which returns a borrowed reference
+#if !defined(PYPY_VERSION) && ((!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030D0000) || (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030D0000))
+    return PyEval_GetFrameGlobals();
+#else
+    PyObject *result = PyEval_GetGlobals();
+    Py_XINCREF(result);
+    return result;
+#endif
+}
+
+PyObject *PepEval_GetFrameBuiltins()
+{
+    // PepEval_GetFrameBuiltins() (added to stable ABI in 3.13) returns a new reference
+    // as opposed to deprecated PyEval_GetBuiltins() which returns a borrowed reference
+#if !defined(PYPY_VERSION) && ((!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030D0000) || (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030D0000))
+    return PyEval_GetFrameBuiltins();
+#else
+    PyObject *result = PyEval_GetBuiltins();
+    Py_XINCREF(result);
+    return result;
+#endif
+}
+
+int PepModule_AddType(PyObject *module, PyTypeObject *type)
+{
+    // PyModule_AddType (added to stable ABI in 3.10) is the replacement for
+    // PyModule_AddObject() (deprecated in 3.13) for adding types to a module.
+#if !defined(PYPY_VERSION) && ((!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030A0000) || (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030A0000))
+    return PyModule_AddType(module, type);
+#else
+    auto *ob = reinterpret_cast<PyObject *>(type);
+    int result = PyModule_AddObject(module, PepType_GetNameStr(type), ob);
+    if (result != 0)
+        Py_XDECREF(ob);
+    return result;
+#endif
+}
+
+int PepModule_Add(PyObject *module, const char *name, PyObject *value)
+{
+    // PyModule_Add (added to stable ABI in 3.13) is the replacement for PyModule_AddObject()
+    // (deprecated in 3.13).
+#if !defined(PYPY_VERSION) && ((!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030D0000) || (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030D0000))
+    return PyModule_Add(module, name, value);
+#else
+    int result = PyModule_AddObject(module, name, value);
+    if (result != 0)
+        Py_XDECREF(value);
+    return result;
+#endif
 }
 
 /***************************************************************************
@@ -1074,7 +1283,6 @@ static inline void *PepType_ExTP(PyTypeObject *type, size_t size)
 void
 Pep384_Init()
 {
-    init_PepRuntime();
 #ifdef Py_LIMITED_API
     check_PyTypeObject_valid();
     Pep_GetVerboseFlag();
@@ -1088,3 +1296,17 @@ Pep384_Init()
 }
 
 } // extern "C"
+
+LIBSHIBOKEN_API bool PepExt_Weakref_IsAlive(PyObject *weakRef)
+{
+#if defined(Py_LIMITED_API) || PY_VERSION_HEX < 0x030D0000
+    return PyWeakref_GetObject(weakRef) != Py_None;
+#else
+    // FIXME: Make this the default code path once Limited API has been raised to 3.13
+    // Note: PyWeakref_GetObject() will be removed in 3.15.
+    PyObject *pobj = nullptr;
+    const bool result = PyWeakref_GetRef(weakRef, &pobj) == 1;
+    Py_XDECREF(pobj);
+    return result;
+#endif
+}
