@@ -1,5 +1,6 @@
 # Copyright (C) 2022 The Qt Company Ltd.
 # SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
+from __future__ import annotations
 
 # flake8: noqa E:203
 
@@ -17,21 +18,33 @@ import os
 import struct
 import sys
 import typing
+import collections
 
 from pathlib import Path
 from typing import TypeVar, Generic
 from _imp import is_builtin
 
 
-class ellipsis(object):
+class ellipsis:
     def __repr__(self):
         return "..."
 
 
+# PYSIDE-3135
+# Using _Self to avoid adding the parameter to typing in versions <3.11
+# Drop this when Python 3.11 is the minimum version
+if not hasattr(typing, "Self"):
+    @typing._SpecialForm
+    def Self(self, parameters):
+        raise TypeError(f"{self} is not subscriptable")
+    typing._Self = Self
+else:
+    typing._Self = typing.Self
+
 ellipsis = ellipsis()
 Point = typing.Tuple[int, int]
 Variant = typing.Any
-QImageCleanupFunction = typing.Callable
+QImageCleanupFunction = collections.abc.Callable[..., typing.Any]
 
 # unfortunately, typing.Optional[t] expands to typing.Union[t, NoneType]
 # Until we can force it to create Optional[t] again, we use this.
@@ -40,7 +53,7 @@ NoneType = type(None)
 # PYSIDE-2517: findChild/findChildren type hints:
 # Placeholder so it does not trigger an UNDEFINED error while building.
 # Later it will be bound to a QObject, within the QtCore types extensions
-PlaceHolderType = TypeVar("PlaceHolderType")
+PlaceholderType = TypeVar("PlaceholderType")
 
 _S = TypeVar("_S")
 
@@ -84,12 +97,7 @@ class _NotCalled(str):
         return eval(text, namespace)
 
 
-USE_PEP563 = False
-# Note: we cannot know if this feature has been imported.
-# Otherwise it would be "sys.version_info[:2] >= (3, 7)".
-# We *can* eventually inspect sys.modules and look if
-# the calling module has this future statement set,
-# but should we do that?
+USE_PEP563 = sys.version_info[:2] >= (3, 10)
 
 
 # Some types are abstract. They just show their name.
@@ -119,8 +127,12 @@ class Instance(_NotCalled):
     pass
 
 
+class KeywordOnly(_NotCalled):
+    pass
+
+
 # Parameterized primitive variables
-class _Parameterized(object):
+class _Parameterized:
     def __init__(self, type):
         self.type = type
         self.__name__ = self.__class__.__name__
@@ -143,7 +155,7 @@ class ArrayLikeVariable(_Parameterized):
 StringList = ArrayLikeVariable(str)
 
 
-class Reloader(object):
+class Reloader:
     """
     Reloder class
 
@@ -186,6 +198,10 @@ class Reloader(object):
                 # Modules are in place, we can update the type_map.
                 g.update(g.pop(proc_name)())
 
+            # Also record an efficient list of modules for PySide.
+            if mod_name.startswith("PySide6."):
+                pyside_modules.add(mod_name)
+
 
 def check_module(mod):
     # During a build, there exist the modules already as directories,
@@ -199,7 +215,9 @@ def check_module(mod):
 
 update_mapping = Reloader().update
 type_map = {}
+type_map_tuple = {}
 namespace = globals()  # our module's __dict__
+pyside_modules: set[str] = set()
 
 type_map.update({
     "...": ellipsis,
@@ -211,16 +229,16 @@ type_map.update({
     "int": int,
     "List": ArrayLikeVariable,
     "Optional": typing.Optional,
-    "Iterable": typing.Iterable,
+    "Iterable": collections.abc.Iterable,
     "long": int,
     "long long": int,
     "nullptr": None,
-    "PyCallable": typing.Callable,
+    "PyCallable": collections.abc.Callable[..., typing.Any],
     "PyObject": object,
     "PyObject*": object,
-    "PyArrayObject": ArrayLikeVariable, # numpy
+    "PyArrayObject": ArrayLikeVariable(typing.Any),  # numpy
     "PyPathLike": typing.Union[str, bytes, os.PathLike[str]],
-    "PySequence": typing.Iterable,  # important for numpy
+    "PySequence": collections.abc.Iterable,  # important for numpy
     "PyTypeObject": type,
     "QChar": str,
     "QHash": typing.Dict,
@@ -248,6 +266,7 @@ type_map.update({
     "QSet": typing.Set,
     "QString": str,
     "QLatin1String": str,
+    "QAnyStringView": str,
     "QStringView": str,
     "QStringList": StringList,
     "quint16": int,
@@ -269,6 +288,8 @@ type_map.update({
     "short": int,
     "signed char": int,
     "signed long": int,
+    "std.chrono.milliseconds": int,
+    "std.chrono.seconds": int,
     "std.list": typing.List,
     "std.map": typing.Dict,
     "std.nullptr_t": NoneType,
@@ -296,18 +317,10 @@ type_map.update({
     "ushort": int,
     "void": int,  # be more specific?
     "WId": WId,
-    "zero(bytes)": b"",
-    "zero(Char)": 0,
-    "zero(float)": 0,
-    "zero(int)": 0,
-    "zero(object)": None,
-    "zero(str)": "",
-    "zero(typing.Any)": None,
-    "zero(Any)": None,
     # This can be refined by importing numpy.typing optionally, but better than nothing.
     "numpy.ndarray": typing.List[typing.Any],
     "std.array[int, 4]": typing.List[int],
-    "std.array[float, 4]": typing.List[float]
+    "std.array[float, 4]": typing.List[float],
 })
 
 type_map.update({
@@ -342,6 +355,7 @@ type_map.update({
     "quint8*"       : bytearray,  # only for QCborStreamReader and QCborValue
     "uchar*"        : typing.Union[bytes, bytearray, memoryview],
     "unsigned char*": typing.Union[bytes, bytearray, memoryview],
+    "char*const"    : typing.Union[bytes, bytearray, memoryview],
 })
 
 type_map.update({
@@ -384,8 +398,12 @@ type_map.update({
 # PYSIDE-1538: We need to treat "std::optional" accordingly.
 type_map.update({
     "std.optional": typing.Optional,
-    })
+})
 
+# PYSIDE-2846: A special keyord only switching token.
+type_map.update({
+    "KeywordOnly": KeywordOnly("None"),
+})
 
 # The Shiboken Part
 def init_Shiboken():
@@ -468,15 +486,9 @@ def init_smart():
 
 # The PySide Part
 def init_PySide6_QtCore():
-    from PySide6.QtCore import Qt, QUrl, QDir, QKeyCombination, QObject
-    from PySide6.QtCore import QRect, QRectF, QSize, QPoint, QLocale, QByteArray
+    from PySide6.QtCore import Qt, QUrl, QDir, QByteArray
+    from PySide6.QtCore import QRect, QRectF, QSize, QPoint
     from PySide6.QtCore import QMarginsF  # 5.9
-    from PySide6.QtCore import SignalInstance
-    try:
-        # seems to be not generated by 5.9 ATM.
-        from PySide6.QtCore import Connection
-    except ImportError:
-        pass
 
     type_map.update({
         "' '": " ",
@@ -493,23 +505,21 @@ def init_PySide6_QtCore():
         "NULL": None,  # 5.6, MSVC
         "nullptr": None,  # 5.9
         # PYSIDE-2517: findChild/findChildren type hints:
-        "PlaceHolderType": typing.TypeVar("PlaceHolderType", bound=QObject),
+        "PlaceholderType": typing.TypeVar("PlaceholderType", bound=PySide6.QtCore.QObject),
         "PyBuffer": typing.Union[bytes, bytearray, memoryview],
         "PyByteArray": bytearray,
         "PyBytes": typing.Union[bytes, bytearray, memoryview],
         "PyTuple": typing.Tuple,
+        "QDeadlineTimer.Forever": PySide6.QtCore.QDeadlineTimer.ForeverConstant.Forever,
         "QDeadlineTimer(QDeadlineTimer.Forever)": Instance("PySide6.QtCore.QDeadlineTimer"),
         "PySide6.QtCore.QUrl.ComponentFormattingOptions":
             PySide6.QtCore.QUrl.ComponentFormattingOption,  # mismatch option/enum, why???
         "PyUnicode": typing.Text,
-        "QByteArrayView": QByteArray,
+        "QByteArray": typing.Union[PySide6.QtCore.QByteArray, bytes, bytearray, memoryview],
+        "QByteArrayView": PySide6.QtCore.QByteArray,
         "Q_NULLPTR": None,
         "QCalendar.Unspecified": PySide6.QtCore.QCalendar.Unspecified,
         "QCborTag(-1)": ulong_max,
-        "QDir.Filters(AllEntries | NoDotAndDotDot)": Instance(
-            "QDir.Filters(QDir.AllEntries | QDir.NoDotAndDotDot)"),
-        "QDir.SortFlags(Name | IgnoreCase)": Instance(
-            "QDir.SortFlags(QDir.Name | QDir.IgnoreCase)"),
         "QEvent.Type.None": None,
         "QGenericArgument((0))": ellipsis,  # 5.6, RHEL 6.6. Is that ok?
         "QGenericArgument()": ellipsis,
@@ -520,32 +530,39 @@ def init_PySide6_QtCore():
         "QJsonObject": typing.Dict[str, PySide6.QtCore.QJsonValue],
         "QModelIndex()": Invalid("PySide6.QtCore.QModelIndex"),  # repr is btw. very wrong, fix it?!
         "QModelIndexList": typing.List[PySide6.QtCore.QModelIndex],
-        "PySideSignalInstance": SignalInstance,
+        "PySideSignalInstance": PySide6.QtCore.SignalInstance,
         "QString()": "",
         "Flag.Default": Instance("PySide6.QtCore.QStringConverterBase.Flags"),
         "QStringList()": [],
         "QStringRef": str,
         "QStringRef": str,
         "Qt.HANDLE": int,  # be more explicit with some constants?
-        "QUrl.FormattingOptions(PrettyDecoded)": Instance(
-            "QUrl.FormattingOptions(QUrl.PrettyDecoded)"),
+        "QUrl.FormattingOptions(QUrl.ComponentFormattingOption.PrettyDecoded)":
+            PySide6.QtCore.QUrl.ComponentFormattingOption.PrettyDecoded,
         "QVariant()": Invalid(Variant),
         "QVariant.Type": type,  # not so sure here...
         "QVariantMap": typing.Dict[str, Variant],
         "std.chrono.seconds{5}" : ellipsis,
+        "Internal.defaultTryTimeout": 5000,
+        "static_cast<int>(Internal.defaultTryTimeout.count())": 5000
     })
-    try:
+    from shibokensupport.signature.parser import using_snake_case
+    if using_snake_case():
         type_map.update({
-            "PySide6.QtCore.QMetaObject.Connection": PySide6.QtCore.Connection,  # wrong!
+            "QKeyCombination.fromCombined(0)": PySide6.QtCore.QKeyCombination.from_combined(0),
         })
-    except AttributeError:
-        # this does not exist on 5.9 ATM.
-        pass
-
     # special case - char* can either be 'bytes' or 'str'. The default is 'bytes'.
     # Here we manually set it to map to 'str'.
-    type_map.update({("PySide6.QtCore.QObject.setProperty", "char*"): str})
-    type_map.update({("PySide6.QtCore.QObject.property", "char*"): str})
+    type_map_tuple.update({("PySide6.QtCore.QObject.setProperty", "char*"): str})
+    type_map_tuple.update({("PySide6.QtCore.QObject.property", "char*"): str})
+    type_map_tuple.update({("PySide6.QtCore.QObject.inherits", "char*"): str})
+    type_map_tuple.update({("PySide6.QtCore.QObject.connect", "char*"): str})
+    type_map_tuple.update({("PySide6.QtCore.QObject.disconnect", "char*"): str})
+    type_map_tuple.update({("PySide6.QtCore.QObject.receivers", "char*"): str})
+    type_map_tuple.update({("PySide6.QtCore.qtTrId", "char*"): str})
+    # special case - char default is 'int'.
+    # Here we manually set it to map to 'str'.
+    type_map_tuple.update({("PySide6.QtCore.QLocale.toString", "char"): str})
 
     return locals()
 
@@ -582,7 +599,7 @@ def init_PySide6_QtGui():
 
     # special case - char* can either be 'bytes' or 'str'. The default is 'bytes'.
     # Here we manually set it to map to 'str'.
-    type_map.update({("PySide6.QtGui.QPixmap.save", "char*"): str})
+    type_map_tuple.update({("PySide6.QtGui.QPixmap.save", "char*"): str})
 
     return locals()
 
@@ -610,7 +627,7 @@ def init_PySide6_QtWidgets():
 def init_PySide6_QtSql():
     from PySide6.QtSql import QSqlDatabase
     type_map.update({
-        "QLatin1StringView(QSqlDatabase.defaultConnection)": QSqlDatabase.defaultConnection,
+        "QSqlDatabase.defaultConnectionName()": "",
         "QVariant.Invalid": Invalid("Variant"),  # not sure what I should create, here...
     })
     return locals()
@@ -659,9 +676,7 @@ def init_PySide6_QtQuick():
 
 
 def init_PySide6_QtTest():
-    from PySide6.QtCore import SignalInstance
     type_map.update({
-        "PySideSignalInstance": SignalInstance,
         "PySide6.QtTest.QTest.PySideQTouchEventSequence": PySide6.QtTest.QTest.QTouchEventSequence,
         "PySide6.QtTest.QTouchEventSequence": PySide6.QtTest.QTest.QTouchEventSequence,
     })
@@ -725,4 +740,43 @@ def init_testbinding():
     })
     return locals()
 
+
+# Functions which should return Optional(result) but don't.
+missing_optional_return = {
+    "PySide6.QtGui.QGuiApplication.modalWindow",
+    "PySide6.QtGui.QGuiApplication.screenAt",
+    "PySide6.QtWidgets.QApplication.activeModalWidget",
+    "PySide6.QtWidgets.QApplication.activePopupWidget",
+    "PySide6.QtWidgets.QApplication.activeWindow",
+    "PySide6.QtWidgets.QApplication.focusWidget",
+    "PySide6.QtWidgets.QApplication.setStyle",
+    "PySide6.QtWidgets.QApplication.topLevelAt",
+    "PySide6.QtWidgets.QApplication.widgetAt",
+    "PySide6.QtWidgets.QComboBox.completer",
+    "PySide6.QtWidgets.QComboBox.lineEdit",
+    "PySide6.QtWidgets.QComboBox.validator",
+    "PySide6.QtWidgets.QCompleter.popup",
+    "PySide6.QtWidgets.QGridLayout.itemAt",
+    "PySide6.QtWidgets.QGridLayout.itemAtPosition",
+    "PySide6.QtWidgets.QLayout.itemAt",
+    "PySide6.QtWidgets.QTableWidget.horizontalHeaderItem",
+    "PySide6.QtWidgets.QTableWidget.item",
+    "PySide6.QtWidgets.QTableWidget.itemAt",
+    "PySide6.QtWidgets.QTableWidget.mimeData",
+    "PySide6.QtWidgets.QTreeWidget.takeTopLevelItem",
+    "PySide6.QtWidgets.QTreeWidget.topLevelItem",
+    "PySide6.QtWidgets.QWidget.childAt",
+    "PySide6.QtWidgets.QWidget.find",
+    "PySide6.QtWidgets.QWidget.focusProxy",
+    "PySide6.QtWidgets.QWidget.graphicsEffect",
+    "PySide6.QtWidgets.QWidget.graphicsProxyWidget",
+    "PySide6.QtWidgets.QWidget.keyboardGrabber",
+    "PySide6.QtWidgets.QWidget.layout",
+    "PySide6.QtWidgets.QWidget.mouseGrabber",
+    "PySide6.QtWidgets.QWidget.nativeParentWidget",
+    "PySide6.QtWidgets.QWidget.nextInFocusChain",
+    "PySide6.QtWidgets.QWidget.parentWidget",
+    "PySide6.QtWidgets.QWidget.previousInFocusChain",
+}
+
 # end of file
